@@ -25,6 +25,9 @@ use std::future::Future;
 use std::str::FromStr;
 use std::time::SystemTime;
 
+#[cfg(test)]
+mod negative_tests;
+
 /// Secret key used to produce a signature which proves that an
 /// input's witness stack contains transaction data equivalent to the
 /// transaction which includes the input itself.
@@ -482,9 +485,9 @@ pub struct Borrower0 {
 
 impl Borrower0 {
     #[allow(clippy::too_many_arguments)]
-    pub async fn new<R, CS, CF>(
+    pub async fn new<R>(
         rng: &mut R,
-        coin_selector: CS,
+        collateral_inputs: Vec<Input>,
         address: Address,
         address_blinding_sk: SecretKey,
         collateral_amount: Amount,
@@ -494,11 +497,8 @@ impl Borrower0 {
     ) -> Result<Self>
     where
         R: RngCore + CryptoRng,
-        CS: FnOnce(Amount, AssetId) -> CF,
-        CF: Future<Output = Result<Vec<Input>>>,
     {
         let keypair = make_keypair(rng);
-        let collateral_inputs = coin_selector(collateral_amount, bitcoin_asset_id).await?;
 
         Ok(Self {
             keypair,
@@ -536,17 +536,20 @@ impl Borrower0 {
     {
         let transaction = loan_response.transaction;
 
-        let principal_amount = transaction
-            .output
-            .iter()
-            .find_map(|out| {
+        let principal_amount = {
+            let expected_output = transaction.output.iter().find_map(|out| {
                 let unblinded_out = out.unblind(secp, self.address_blinding_sk).ok()?;
                 let is_principal_out = unblinded_out.asset == self.usdt_asset_id
                     && out.script_pubkey == self.address.script_pubkey();
 
-                is_principal_out.then(|| Amount::from_sat(unblinded_out.value))
-            })
-            .context("no principal txout")?;
+                is_principal_out.then(|| unblinded_out)
+            });
+
+            match expected_output {
+                Some(out) => Amount::from_sat(out.value),
+                None => bail!("could not find principal output"),
+            }
+        };
 
         let collateral_contract = loan_response.collateral_contract;
         let collateral_address = collateral_contract.address();
@@ -554,18 +557,18 @@ impl Borrower0 {
         let collateral_blinding_sk = loan_response.repayment_collateral_input.blinding_key;
         let collateral_script_pubkey = collateral_address.script_pubkey();
 
-        transaction
-            .output
-            .iter()
-            .find_map(|out| {
-                let unblinded_out = out.unblind(secp, collateral_blinding_sk).ok()?;
-                let is_collateral_out = unblinded_out.asset == self.bitcoin_asset_id
-                    && unblinded_out.value == self.collateral_amount.as_sat()
-                    && out.script_pubkey == collateral_script_pubkey;
+        let collateral_output = transaction.output.iter().find_map(|out| {
+            let unblinded_out = out.unblind(secp, collateral_blinding_sk).ok()?;
+            let is_collateral_out = unblinded_out.asset == self.bitcoin_asset_id
+                && unblinded_out.value == self.collateral_amount.as_sat()
+                && out.script_pubkey == collateral_script_pubkey;
 
-                is_collateral_out.then(|| out)
-            })
-            .context("no collateral txout")?;
+            is_collateral_out.then(|| out)
+        });
+
+        if collateral_output.is_none() {
+            bail!("could not find collateral output")
+        }
 
         let collateral_input_amount = self
             .collateral_inputs
@@ -590,18 +593,23 @@ impl Borrower0 {
                 )
             })?;
 
-        transaction
-            .output
-            .iter()
-            .find_map(|out| {
+        if collateral_change_amount != Amount::ZERO {
+            let expected_output = transaction.output.iter().find_map(|out| {
                 let unblinded_out = out.unblind(secp, self.address_blinding_sk).ok()?;
                 let is_collateral_change_out = unblinded_out.asset == self.bitcoin_asset_id
                     && unblinded_out.value == collateral_change_amount.as_sat()
                     && out.script_pubkey == self.address.script_pubkey();
 
                 is_collateral_change_out.then(|| out)
-            })
-            .context("no collateral change txout")?;
+            });
+
+            if expected_output.is_none() {
+                bail!(
+                    "could not find collateral change output with amount {}",
+                    collateral_change_amount
+                )
+            }
+        }
 
         Ok(Borrower1 {
             keypair: self.keypair,
@@ -914,14 +922,14 @@ impl Lender0 {
     /// Interpret a loan request and performs lender logic.
     ///
     /// rate is expressed in usdt sats per btc, i.e. rate = 1 BTC / USDT
-    #[deprecated(note = "Use interpret_loan_request instead", since = "0.3.0")]
+    #[deprecated(note = "Use build_loan_transaction instead", since = "0.3.0")]
     pub async fn interpret<R, C, CS, CF>(
         self,
         rng: &mut R,
         secp: &Secp256k1<C>,
         coin_selector: CS,
         loan_request: LoanRequest,
-        timelock: u32,
+        timelock: Timelock,
         rate: u64,
     ) -> Result<Lender1>
     where
@@ -966,13 +974,14 @@ impl Lender0 {
         repayment_amount: Amount,
         min_collateral_price: u64,
         (borrower_pk, borrower_address): (PublicKey, Address),
-        timelock: u32,
+        timelock: impl Into<Timelock>,
     ) -> Result<Lender1>
     where
         R: RngCore + CryptoRng,
         C: Verification + Signing,
     {
         let chain = Chain::new(&borrower_address, &self.address)?;
+        let timelock = Into::<Timelock>::into(timelock);
 
         let collateral_inputs = collateral_inputs
             .into_iter()
@@ -994,6 +1003,10 @@ impl Lender0 {
             .collect::<Vec<_>>();
 
         let inputs = borrower_inputs.chain(lender_inputs).collect::<Vec<_>>();
+        let inputs_not_last_confidential = inputs
+            .iter()
+            .map(|(asset, secrets)| (*asset, Some(*secrets)))
+            .collect::<Vec<_>>();
 
         let collateral_input_amount = collateral_inputs
             .iter()
@@ -1028,33 +1041,6 @@ impl Lender0 {
         let now = std::time::SystemTime::now().duration_since(SystemTime::UNIX_EPOCH)?;
         let contract_creation_timestamp = now.as_secs() + 300;
 
-        let collateral_contract = CollateralContract::new(
-            borrower_pk,
-            lender_pk,
-            timelock,
-            (repayment_principal_output, self.address_blinder),
-            self.oracle_pk,
-            min_collateral_price,
-            contract_creation_timestamp,
-            chain,
-        )
-        .context("could not build collateral contract")?;
-        let collateral_address = collateral_contract.blinded_address(collateral_blinding_pk.key);
-
-        let inputs_not_last_confidential = inputs
-            .iter()
-            .map(|(asset, secrets)| (*asset, Some(*secrets)))
-            .collect::<Vec<_>>();
-        let (collateral_tx_out, abf_collateral, vbf_collateral) = TxOut::new_not_last_confidential(
-            rng,
-            secp,
-            collateral_amount.as_sat(),
-            collateral_address.clone(),
-            self.bitcoin_asset_id,
-            inputs_not_last_confidential.as_slice(),
-        )
-        .context("could not construct collateral txout")?;
-
         let (principal_tx_out, abf_principal, vbf_principal) = TxOut::new_not_last_confidential(
             rng,
             secp,
@@ -1065,45 +1051,15 @@ impl Lender0 {
         )
         .context("could not construct principal txout")?;
 
+        let tx_fee = Amount::from_sat(
+            estimate_virtual_size(inputs.len() as u64, 4) * fee_sats_per_vbyte.as_sat(),
+        );
+
         let principal_input_amount = unblinded_principal_inputs
             .iter()
             .fold(0, |sum, input| sum + input.secrets.value);
         let principal_change_amount = Amount::from_sat(principal_input_amount) - principal_amount;
-        let (principal_change_tx_out, abf_principal_change, vbf_principal_change) =
-            TxOut::new_not_last_confidential(
-                rng,
-                secp,
-                principal_change_amount.as_sat(),
-                self.address.clone(),
-                self.usdt_asset_id,
-                &inputs_not_last_confidential,
-            )
-            .context("could not construct principal change txout")?;
 
-        let not_last_confidential_outputs = [
-            &TxOutSecrets::new(
-                self.bitcoin_asset_id,
-                abf_collateral,
-                collateral_amount.as_sat(),
-                vbf_collateral,
-            ),
-            &TxOutSecrets::new(
-                self.usdt_asset_id,
-                abf_principal,
-                principal_amount.as_sat(),
-                vbf_principal,
-            ),
-            &TxOutSecrets::new(
-                self.usdt_asset_id,
-                abf_principal_change,
-                principal_change_amount.as_sat(),
-                vbf_principal_change,
-            ),
-        ];
-
-        let tx_fee = Amount::from_sat(
-            estimate_virtual_size(inputs.len() as u64, 4) * fee_sats_per_vbyte.as_sat(),
-        );
         let collateral_change_amount = Amount::from_sat(collateral_input_amount)
             .checked_sub(collateral_amount)
             .map(|a| a.checked_sub(tx_fee))
@@ -1114,20 +1070,108 @@ impl Lender0 {
                     collateral_amount, tx_fee, collateral_input_amount,
                 )
             })?;
-        let (collateral_change_tx_out, _, _) = TxOut::new_last_confidential(
+
+        let mut tx_outs = vec![principal_tx_out];
+        let mut before_last_confidential_output_secrets = vec![TxOutSecrets::new(
+            self.usdt_asset_id,
+            abf_principal,
+            principal_amount.as_sat(),
+            vbf_principal,
+        )];
+
+        if collateral_change_amount == Amount::ZERO {
+            log::info!(
+                "No change output needed for collateral asset: {}",
+                self.bitcoin_asset_id
+            );
+        } else {
+            let (tx_out, abf, vbf) = TxOut::new_not_last_confidential(
+                rng,
+                secp,
+                collateral_change_amount.as_sat(),
+                borrower_address,
+                self.bitcoin_asset_id,
+                &inputs_not_last_confidential,
+            )
+            .with_context(|| {
+                format!(
+                    "could not construct change output for collateral asset: {}",
+                    self.bitcoin_asset_id
+                )
+            })?;
+
+            tx_outs.push(tx_out);
+            before_last_confidential_output_secrets.push(TxOutSecrets::new(
+                self.bitcoin_asset_id,
+                abf,
+                collateral_change_amount.as_sat(),
+                vbf,
+            ))
+        }
+
+        if principal_change_amount == Amount::ZERO {
+            log::info!(
+                "No change output needed for principal asset: {}",
+                self.usdt_asset_id
+            );
+        } else {
+            let (tx_out, abf, vbf) = TxOut::new_not_last_confidential(
+                rng,
+                secp,
+                principal_change_amount.as_sat(),
+                self.address.clone(),
+                self.usdt_asset_id,
+                &inputs_not_last_confidential,
+            )
+            .with_context(|| {
+                format!(
+                    "could not construct change output for principal asset: {}",
+                    self.usdt_asset_id
+                )
+            })?;
+
+            tx_outs.push(tx_out);
+            before_last_confidential_output_secrets.push(TxOutSecrets::new(
+                self.usdt_asset_id,
+                abf,
+                principal_change_amount.as_sat(),
+                vbf,
+            ))
+        }
+
+        let collateral_contract = CollateralContract::new(
+            borrower_pk,
+            lender_pk,
+            timelock.into(),
+            (repayment_principal_output, self.address_blinder),
+            self.oracle_pk,
+            min_collateral_price,
+            contract_creation_timestamp,
+            chain,
+        )
+        .context("could not build collateral contract")?;
+        let collateral_address = collateral_contract.blinded_address(collateral_blinding_pk.key);
+        let (collateral_tx_out, _, _) = TxOut::new_last_confidential(
             rng,
             secp,
-            collateral_change_amount.as_sat(),
-            borrower_address,
+            collateral_amount.as_sat(),
+            collateral_address.clone(),
             self.bitcoin_asset_id,
             inputs
                 .iter()
                 .map(|(asset, secrets)| (*asset, *secrets))
                 .collect::<Vec<_>>()
                 .as_slice(),
-            &not_last_confidential_outputs,
+            &before_last_confidential_output_secrets
+                .iter()
+                .collect::<Vec<_>>(),
         )
-        .context("Creation of collateral change output failed")?;
+        .context("could not construct collateral txout")?;
+
+        tx_outs.push(collateral_tx_out.clone());
+
+        let tx_fee_tx_out = TxOut::new_fee(tx_fee.as_sat(), self.bitcoin_asset_id);
+        tx_outs.push(tx_fee_tx_out);
 
         let tx_ins = {
             let borrower_inputs = collateral_inputs.iter().map(|input| input.txin);
@@ -1146,19 +1190,11 @@ impl Lender0 {
                 .collect::<Vec<_>>()
         };
 
-        let tx_fee_tx_out = TxOut::new_fee(tx_fee.as_sat(), self.bitcoin_asset_id);
-
         let loan_transaction = Transaction {
             version: 2,
             lock_time: 0,
             input: tx_ins,
-            output: vec![
-                collateral_tx_out.clone(),
-                principal_tx_out,
-                principal_change_tx_out,
-                collateral_change_tx_out,
-                tx_fee_tx_out,
-            ],
+            output: tx_outs,
         };
 
         let repayment_collateral_input = {
@@ -1257,7 +1293,7 @@ struct LoanAmounts {
 pub struct Lender1 {
     keypair: (SecretKey, PublicKey),
     address: Address,
-    timelock: u32,
+    timelock: Timelock,
     loan_transaction: Transaction,
     collateral_contract: CollateralContract,
     collateral_amount: Amount,
@@ -1343,7 +1379,7 @@ impl Lender1 {
 
         let mut liquidation_transaction = Transaction {
             version: 2,
-            lock_time: self.timelock,
+            lock_time: self.timelock.into(),
             input: tx_ins,
             output: tx_outs,
         };
@@ -1473,6 +1509,61 @@ pub mod transaction_as_string {
     }
 }
 
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub enum Timelock {
+    Timestamp(u32),
+    BlockHeight(u32),
+}
+
+impl Timelock {
+    // https://github.com/bitcoin/bitcoin/blob/b620b2d58a55a88ad21da70cb2000863ef17b651/src/script/script.h#L37-L39
+    const LOCKTIME_THRESHOLD: u32 = 500000000;
+
+    pub fn new_timestamp(n: u32) -> Result<Self, NotATimestamp> {
+        if n < Self::LOCKTIME_THRESHOLD {
+            return Err(NotATimestamp(n));
+        }
+
+        Ok(Self::Timestamp(n))
+    }
+
+    pub fn new_block_height(n: u32) -> Result<Self, NotABlockHeight> {
+        if n >= Self::LOCKTIME_THRESHOLD {
+            return Err(NotABlockHeight(n));
+        }
+
+        Ok(Self::BlockHeight(n))
+    }
+}
+
+#[derive(thiserror::Error, Debug, PartialEq)]
+#[error("Timelock based on timestamp must be over 500000000, got {0}")]
+pub struct NotATimestamp(u32);
+
+#[derive(thiserror::Error, Debug, PartialEq)]
+#[error("Timelock based on block height must be under 500000000, got {0}")]
+pub struct NotABlockHeight(u32);
+
+impl From<Timelock> for u32 {
+    fn from(timelock: Timelock) -> Self {
+        match timelock {
+            Timelock::Timestamp(inner) | Timelock::BlockHeight(inner) => inner,
+        }
+    }
+}
+
+impl From<u32> for Timelock {
+    fn from(n: u32) -> Self {
+        log::warn!("Choose a Timelock type explicitly via constructors");
+
+        if n < Timelock::LOCKTIME_THRESHOLD {
+            Self::BlockHeight(n)
+        } else {
+            Self::Timestamp(n)
+        }
+    }
+}
+
 /// Possible networks on which the loan contract may be deployed.
 #[derive(Debug, Clone, Copy, serde::Serialize, serde::Deserialize)]
 enum Chain {
@@ -1552,3 +1643,25 @@ mod constant_tests {
         assert_eq!(actual, expected);
     }
 }
+
+#[cfg(test)]
+mod timelock_tests {
+    use super::{NotABlockHeight, NotATimestamp, Timelock};
+    use proptest::prelude::*;
+
+    proptest! {
+        #[test]
+        fn locktime_under_threshold_is_blocktime(n in 0u32..=Timelock::LOCKTIME_THRESHOLD) {
+            prop_assert_eq!(Timelock::new_block_height(n), Ok(Timelock::BlockHeight(n)));
+            prop_assert_eq!(Timelock::new_timestamp(n), Err(NotATimestamp(n)));
+        }
+    }
+
+    proptest! {
+        #[test]
+        fn locktime_over_threshold_is_timestamp(n in Timelock::LOCKTIME_THRESHOLD..=u32::MAX) {
+            prop_assert_eq!(Timelock::new_timestamp(n), Ok(Timelock::Timestamp(n)));
+            prop_assert_eq!(Timelock::new_block_height(n), Err(NotABlockHeight(n)));
+        }
+    }
+}
